/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
**+
** NAME:
**	xfpovray
**
** PURPOSE:
**      An XForms interface to the Persistence of Vision Raytracer v3
**
**      The basic method is as follows:  Options are set on the widget; options
**      not set are assigned default values.  All options are written out to
**      a file that is read by POV-ray upon startup.
**
**      User options in xfpovray.h 
**
**
** MODIFICATION HISTORY:
**	v0.9: 1996 July, written, mallozzi@gibson.msfc.nasa.gov
**                       <mallozzir@cspar.uah.edu>
**
**      v1.0: 1996 July, RSM 
**                       o Added "Abort Render" button
**                       o Added support for Mosaic Preview     
**-
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include "xfpovray.h"
#include "forms.h"
 
/* POV-ray defaults */
#define TIME 2
#define PUSHED 1
#define RELEASED 0
#define POV_DISPLAY_DURING_RENDER 1
#define POV_OUTPUT_TO_FILE 1
#define POV_VERBOSE 1
#define POV_ANTIALIAS 0
#define POV_ANTIALIAS_THRESHOLD 0.3
#define POV_AA_SAMPLING_METHOD 1
#define POV_AA_DEPTH 3
#define POV_RADIOSITY 0
#define POV_QUALITY 9
#define POV_REDIRECT_ON 0
#define POV_VERSION 3
#define POV_REMOVE_BOUNDS 1
#define POV_SPLIT_UNIONS 0
#define POV_BOUNDING 1
#define POV_BOUNDING_THRESHOLD 25
#define POV_LIGHT_BUFFER 1
#define POV_VISTA_BUFFER 1
#define POV_JITTER 1
#define POV_JITTER_AMOUNT 1.0
#define POV_INITIAL_FRAME 1
#define POV_FINAL_FRAME -1
#define POV_CYCLIC_ANIMATION 0
#define POV_FIELD_RENDERING 0
#define POV_ODD_FIELD 0
#define POV_PALETTE 1
#define POV_OUTPUT_FILETYPE 1
 

/* Shouldn't need to edit below here */

int povray_pid = 0;

/* Geometry for the options widget */
const int GRP_X = 10;
const int GRP_Y = 250;
const int GRP_W = 400;
const int GRP_H = 305;

/* Callback functions */
void QuitCB();
void GetFileCB();
void SetImageSizeCB();
void PovCB();
void RadioCB();
void ConfigCB();
void YesNoCB();
void QualityCB();
void QualityResetCB();
void ClearPartialImageCB(); 
void HelpCB();
void DismissMessageCB();
void DismissAboutCB();
void ContinueCB(); 
void TogglePixmapCB();
void AbortCB();
void SetMosaicCB();
int  MoveCB();
 
/* Other functions */
void write_config();
void ShowImage();

/* The options buttons */
FL_OBJECT *opt[4]; 

/* Option groups appear and disappear; one is always the current group */
FL_OBJECT *current_grp;

/* The actual option groups */
void pixmap_group();
FL_OBJECT *pixmap_grp;

void parsing_group();
FL_OBJECT *parsing_grp;

void output_group();
FL_OBJECT *output_grp;

void tracing_group();
FL_OBJECT *tracing_grp;

void animation_group();
FL_OBJECT *animation_grp;
 
void config_group();
FL_OBJECT *config_grp;

/* Global objects; what a mess, but I never claimed to be an 
** an expert c programmer ;-) 
*/
FL_FORM   *main_form, *text_form, *fatal_form, *about_form;
FL_OBJECT *file_input, *height_input, *width_input;
FL_OBJECT *start_column, *end_column, *start_row, *end_row;
FL_OBJECT *antialias_threshold, *bounding_threshold;
FL_OBJECT *config_browser; 
FL_OBJECT *quality_slider, *aa_slider;
FL_OBJECT *libpath_input[5];
FL_OBJECT *text, *fatal_text; 
FL_OBJECT *clock_input, *initial_clock_input, *final_clock_input;
FL_OBJECT *subset_start_frame_input, *subset_end_frame_input;
FL_OBJECT *init_frame_input, *final_frame_input, *jitter_input;
FL_OBJECT *tc, *tim, *auto_image, *toggle;
FL_OBJECT *mosaic_start, *mosaic_end;

int auto_view = 0;
int loaded_inifile = 0; 

/* The options */
struct pov_options {
       const char *input_file_name;
       int height;
       int width;
       int display_during_render;
       int output_to_file;
       int verbose;
       int antialias;
       int aa_sampling_method;
       int aa_depth;
       int radiosity;
       int quality;
       int redirect_on;
       int version; 
       int remove_bounds;
       int split_unions;
       int bounding;
       int bounding_threshold;
       int light_buffer;
       int vista_buffer;
       int jitter;
       int initial_frame;
       int final_frame;
       int cyclic_animation;
       int field_rendering;
       int odd_field;
       int palette;
       int output_filetype;
};

/* Set options to defaults */
struct pov_options pov = {
       "", /* input_file_name */ 
       POV_HEIGHT,
       POV_WIDTH,
       POV_DISPLAY_DURING_RENDER,
       POV_OUTPUT_TO_FILE,
       POV_VERBOSE,
       POV_ANTIALIAS,
       POV_AA_SAMPLING_METHOD,
       POV_AA_DEPTH,
       POV_RADIOSITY,
       POV_QUALITY,
       POV_REDIRECT_ON,
       POV_VERSION, 
       POV_REMOVE_BOUNDS,
       POV_SPLIT_UNIONS,
       POV_BOUNDING,
       POV_BOUNDING_THRESHOLD,
       POV_LIGHT_BUFFER,
       POV_VISTA_BUFFER,
       POV_JITTER,
       POV_INITIAL_FRAME,
       POV_FINAL_FRAME,
       POV_CYCLIC_ANIMATION,
       POV_FIELD_RENDERING,
       POV_ODD_FIELD,
       POV_PALETTE,
       POV_OUTPUT_FILETYPE
};

/* Temporary output files */
char POV_INIFILE[1024];
char STREAM_FILE[1024];
char FATAL_FILE[1024];



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void main (int argc, char **argv)
{
 
    FL_OBJECT *obj, *ret;
    FL_IOPT control;       
    int version, revision;
    static char *home;
   

    /* Make sure the distribution is consistent wrt version numbers */
    if (FL_INCLUDE_VERSION != fl_library_version(&version, &revision)) {
       fprintf(stderr,"*** Header/library version mismatch!\n");
       fprintf(stderr,"    Header: %d.%d\n", FL_VERSION, FL_REVISION);
       fprintf(stderr,"    Library: %d.%d\n", version, revision);
       exit(1);
    }
    
    /* Get home directory, and set filepaths */
    home = getenv("HOME");
    sprintf(POV_INIFILE, "%s/%s", home, _POV_INIFILE);
    sprintf(STREAM_FILE, "%s/%s", home, _STREAM_FILE);
    sprintf(FATAL_FILE, "%s/%s", home, _FATAL_FILE);
    

    /* Set some defaults and initialize the library */
    control.inputFontSize   = 12;
    control.browserFontSize = 12;
    control.buttonFontSize  = 12;
    control.labelFontSize   = 12;
    control.menuFontSize    = 12;
    fl_set_defaults(FL_PDInputFontSize|FL_PDBrowserFontSize|
        FL_PDButtonFontSize|FL_PDLabelFontSize|
        FL_PDMenuFontSize, &control);
    
    fl_initialize (&argc, argv, "xfpovray", 0, 0);


    main_form = fl_bgn_form(FL_UP_BOX, 420, 580);

        obj = fl_add_box(FL_UP_BOX, 5, 5, 420-10, 25, "");

	obj = fl_add_menu(FL_PUSH_MENU, 
	    10, 5, 30, 25, "File");
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_boxtype(obj, FL_NO_BOX);
            fl_set_menu(obj, "Quit");
            fl_set_object_callback(obj, QuitCB, 0);        

	obj = fl_add_menu(FL_PUSH_MENU, 
	    380, 5, 30, 25, "Help");
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_boxtype(obj, FL_NO_BOX);
            fl_set_menu(obj, "Help|About");        
            fl_set_object_callback(obj, HelpCB, 0);
                                
        obj = fl_add_frame(FL_ENGRAVED_FRAME, 10, 35, 400, 200, "");
  
        /* Left side of top widget */
	obj = fl_add_button(FL_NORMAL_BUTTON, 
	    15, 40, 100, 25, "Scene File"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_callback(obj, GetFileCB, 0);

	file_input = fl_add_input(FL_NORMAL_INPUT, 
	    125, 40, 280, 25, "");
            fl_set_input(file_input, "");  

	obj = fl_add_button(FL_NORMAL_BUTTON, 
	    15, 70, 100, 25, "Edit Scene"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_callback(obj, GetFileCB, 1);

	obj = fl_add_box(FL_FRAME_BOX, 
	    125, 70, 100, 25, POV_EDITOR); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);

	obj = fl_add_button(FL_NORMAL_BUTTON, 
	    15, 100, 100, 25, "View Image"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_callback(obj, GetFileCB, 2);

        auto_image = fl_add_choice(FL_NORMAL_CHOICE, 
            125, 100, 100, 25, ""); 
            fl_addto_choice(auto_image, "Auto OFF|Auto ON");
            fl_set_choice(auto_image, auto_view + 1);
            fl_set_choice_fontstyle(auto_image, FL_BOLD_STYLE); 
            fl_set_object_lstyle(auto_image, FL_BOLD_STYLE);
            fl_set_object_boxtype(auto_image, FL_FRAME_BOX);
            fl_set_object_callback(auto_image, YesNoCB, 4);
            if (pov.display_during_render == 1) {
               fl_set_object_color(auto_image, FL_COL1, FL_INACTIVE_COL); 
               fl_deactivate_object(auto_image);
            }         
          
	obj = fl_add_menu(FL_PUSH_MENU, 
            15, 130, 100, 25, "Config File"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_boxtype(obj, FL_UP_BOX);
            fl_set_menu(obj, "View|Save|Load");        
            fl_set_object_callback(obj, ConfigCB, 0);

	obj = fl_add_button(FL_NORMAL_BUTTON, 
	    15, 160, 100, 55, "Render\nImage"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_callback(obj, PovCB, 0);
            fl_set_object_color(obj, FL_MCOL, FL_COL1);
            
	obj = fl_add_button(FL_NORMAL_BUTTON, 
	    125, 160, 100, 25, "Abort Render"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_callback(obj, AbortCB, 0);
 
 
        /* Right side of top widget */ 
 	toggle = fl_add_button(FL_NORMAL_BUTTON, 
            290, 70, 115, 25, "Options");
            fl_set_object_lstyle(toggle, FL_BOLD_STYLE);
            fl_set_object_boxtype(toggle, FL_NO_BOX);
            fl_set_object_callback(toggle, TogglePixmapCB, 0);

        /* Option radio buttons */
        fl_bgn_group();
        
            opt[0] = fl_add_lightbutton(FL_RADIO_BUTTON, 
                290, 100, 115, 25,
                "Parsing");
            fl_set_object_lstyle(opt[0], FL_BOLD_STYLE);
            fl_set_object_callback(opt[0], RadioCB, 0);
 
            opt[1] = fl_add_lightbutton(FL_RADIO_BUTTON, 
                290, 130, 115, 25,
                "Output");
            fl_set_object_lstyle(opt[1], FL_BOLD_STYLE);
            fl_set_object_callback(opt[1], RadioCB, 0);
 
            opt[2] = fl_add_lightbutton(FL_RADIO_BUTTON, 
                290, 160, 115, 25,
                "Tracing");
            fl_set_object_lstyle(opt[2], FL_BOLD_STYLE);
            fl_set_object_callback(opt[2], RadioCB, 0);
           
            opt[3] = fl_add_lightbutton(FL_RADIO_BUTTON, 
                290, 190, 115, 25,
                "Animation");
            fl_set_object_lstyle(opt[3], FL_BOLD_STYLE);
            fl_set_object_callback(opt[3], RadioCB, 0);
  
        fl_end_group();
 

        /* Frame for the options groups */
        obj = fl_add_frame(FL_ENGRAVED_FRAME, GRP_X, GRP_Y, GRP_W, GRP_H, "");
 
        /* Startup pixmap that is hidden once an option button is selected */
        pixmap_group();
        current_grp = pixmap_grp;
  
        /* Option groups -- build and hide for now */ 
        parsing_group();
        fl_hide_object(parsing_grp);
        output_group();
        fl_hide_object(output_grp);
        tracing_group();
        fl_hide_object(tracing_grp);
        animation_group();
        fl_hide_object(animation_grp);
        config_group();
        fl_hide_object(config_grp);
               
    fl_end_form();
    
    /* Show this form if redirect option is on */
    text_form = fl_bgn_form(FL_UP_BOX, 560, 350);

        text = fl_add_browser(FL_NORMAL_BROWSER, 
            15, 15, 530, 300, ""); 
            
        obj = fl_add_button(FL_NORMAL_BUTTON, 
            245, 320, 70, 25, "Dismiss"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_callback(obj, DismissMessageCB, 0);
        
    fl_end_form(); 

    /* Show this form if there is a fatal error */
    fatal_form = fl_bgn_form(FL_UP_BOX, 560, 350);

        fatal_text = fl_add_browser(FL_NORMAL_BROWSER, 
            15, 15, 530, 300, ""); 
            
        obj = fl_add_button(FL_NORMAL_BUTTON, 
            245, 320, 70, 25, "Dismiss"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_callback(obj, DismissMessageCB, 1);
        
    fl_end_form();
     
     
    about_form = fl_bgn_form(FL_UP_BOX, 400, 270);

        tc = fl_add_box(FL_NO_BOX, 
            150, 30, 120, 40, "xfpovray"); 
            fl_set_object_lsize(tc, FL_HUGE_SIZE);
            fl_set_object_lstyle(tc, FL_BOLDITALIC_STYLE|FL_EMBOSSED_STYLE);
            fl_set_object_lcol(tc, FL_RED);
  
        obj = fl_add_box(FL_NO_BOX, 
            150, 130, 120, 25, "v1.0"); 
            fl_set_object_lsize(obj, FL_MEDIUM_SIZE);
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_lcol(obj, FL_BLACK);

        obj = fl_add_box(FL_NO_BOX, 
            150, 150, 120, 25, "Robert S. Mallozzi"); 
            fl_set_object_lsize(obj, FL_MEDIUM_SIZE);
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_lcol(obj, FL_BLACK);

        obj = fl_add_box(FL_NO_BOX, 
            150, 170, 120, 25, "<mallozzir@cspar.uah.edu>"); 
            fl_set_object_lsize(obj, FL_MEDIUM_SIZE);
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_lcol(obj, FL_BLACK);
                           
        obj = fl_add_button(FL_NORMAL_BUTTON, 175, 220, 70, 25, "Dismiss");
            fl_set_object_callback (obj, DismissAboutCB, 0);
            fl_set_object_lsize(obj, FL_NORMAL_SIZE);
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);

    
        tim = fl_add_timer(FL_HIDDEN_TIMER, 0, 0, 1, 1, "");
        fl_set_timer(tim, TIME);
        fl_set_object_callback(tim, ContinueCB, 0);
            
        fl_set_idle_callback(MoveCB, 0);

    fl_end_form();
    

    fl_show_form(main_form, FL_PLACE_CENTERFREE, FL_FULLBORDER, "xfpovray");

 
    do {
       ret = fl_do_forms();
    } while (ret != NULL);

    fl_hide_form(main_form);
    exit(0);



} /* END */




/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void QuitCB(FL_OBJECT *ob, long data)
{

    fl_finish();
    exit (0);
}
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void AbortCB(FL_OBJECT *ob, long data)
{
    if (povray_pid == 0) 
       return;
    kill(povray_pid, SIGINT);
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void TogglePixmapCB(FL_OBJECT *ob, long data)
{
    int i;

    if ( current_grp != pixmap_grp) {

       /* Release any option buttons that might be selected */
       for (i=0; i<4; i++)
           fl_set_button(opt[i], RELEASED);
             
       fl_hide_object(current_grp);
       current_grp = pixmap_grp;
       fl_show_object(current_grp);

    }
    
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void DismissMessageCB(FL_OBJECT *ob, long data)
{

    switch (data) {
           case 0: fl_hide_form(text_form);
                   break;
           case 1: fl_hide_form(fatal_form);
                   break;
           default: perror("xfpovray: DismissMessageCB");
    }
    

}
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void DismissAboutCB(FL_OBJECT *ob, long data)
{
    fl_hide_form(about_form);
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* Help and About */
void HelpCB(FL_OBJECT *ob, long data)
{
    int submenu;
    

    submenu = fl_get_menu(ob);

    switch (submenu) {
           case 1: fl_set_goodies_font(FL_BOLD_STYLE, 12);
                   fl_show_message("Select a Scene File",
                       "Set Options for the Scene", 
                       "Render Image!");
                   break;
           case 2: fl_show_form(about_form, 
                       FL_PLACE_CENTERFREE, FL_FULLBORDER, "xfpovray");
                   break;
           default: perror("xfpovray: HelpCB");
    } 
    
 
} 
 
 
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* Pops up a file selection box */
void GetFileCB(FL_OBJECT *ob, long data)
{   
    const char *temp;
    char command[256];
         

    switch (data) {
       case 0: /* Scene File button */
            pov.input_file_name = fl_show_fselector("Enter pov filename:", 
                POV_DIR, SCENE_FILTER, "");  
            if (pov.input_file_name == NULL) 
               return;
            else
               fl_set_input(file_input, pov.input_file_name);
            break;
       case 1: /* Edit Scene button */
            temp = fl_show_fselector("Enter filename:", 
                POV_DIR, SCENE_FILTER, "");  
            if (temp == NULL) 
               return;
            else {
               sprintf(command, "%s %s &", POV_EDITOR, temp);
               system(command);
            }
            break;
       case 2: /* View Image button */
            temp = fl_show_fselector("Enter filename:", 
                POV_DIR, IMAGE_FILTER, "");  
            if (temp == NULL) 
               return;
            else {
               sprintf(command, "%s %s &", DISPLAY_COMMAND, temp);
               system(command);            
            }
            break;
        default: perror("xfpovray: GetFileCB");

    } /* endswitch */          

}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* The option radio button callback; sets the current group and shows it */
void RadioCB(FL_OBJECT *ob, long data)
{

    fl_hide_object(current_grp);

    if (fl_get_button(opt[0]) == 1) {
       current_grp = parsing_grp;
    }
       
    if (fl_get_button(opt[1]) == 1) { 
       current_grp = output_grp;
    }

    if (fl_get_button(opt[2]) == 1) { 
       current_grp = tracing_grp;
    }

    if (fl_get_button(opt[3]) == 1) {
       current_grp = animation_grp;
    }
                       
    fl_show_object(current_grp);

}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* Set image height or width */
void SetImageSizeCB(FL_OBJECT *ob, long data)
{
    char temp[10];

    if (data == 0) {
       fl_set_input(height_input, fl_get_menu_text(ob));

       if (strcmp(fl_get_input(height_input), "") == 0) {
	  sprintf(temp, "%d", POV_HEIGHT);
	  fl_set_input(height_input, temp);
	  pov.height = POV_HEIGHT;
       }
       else
	  pov.height = atoi(fl_get_input(height_input));
    }
    if (data == 1) {
       fl_set_input(width_input, fl_get_menu_text(ob));

       if (strcmp(fl_get_input(width_input), "") == 0) {
	  sprintf(temp, "%d", POV_WIDTH);
	  fl_set_input(width_input, temp);
	  pov.width = POV_WIDTH;
       }
       else
	  pov.width = atoi(fl_get_input(width_input)); 
    }
    
}

 

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void SetMosaicCB(FL_OBJECT *ob, long data)
{

    if (data == 0) {
       fl_set_input(mosaic_start, fl_get_menu_text(ob));
    }
    if (data == 1) {
       fl_set_input(mosaic_end, fl_get_menu_text(ob));
    }
   
}
 

 

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* Write the config file, then invoke POV-ray */
void PovCB(FL_OBJECT *ob, long data)
{
    char command[256];
    char temp[256];
    int i, pid;
    FILE *fatal_file;

    if (pov.input_file_name == "") {
       fl_set_goodies_font(FL_BOLD_STYLE, 12);
       if (fl_show_question("You have not entered", "a scene file.",
          "Select scene file now?") == 1) {
        
          pov.input_file_name = fl_show_fselector("Enter pov filename:", 
              POV_DIR, SCENE_FILTER, "");  
          if (pov.input_file_name == NULL) 
             return;
          else
             fl_set_input(file_input, pov.input_file_name);
       }
       else 
          return;
    }       

    if (loaded_inifile != 1) 
       write_config();
    else
       loaded_inifile = 0;

    /* Execute POV-Ray as a child process so we can interrupt a 
    ** render by sending SIGINT
    */
    pid = fork();
    
    if (pid == 0) {
       /* In child process; invoke POV-Ray */

       sprintf(command, "%s", POV_COMMAND);
       if (execlp (command, "xfpovray", POV_INIFILE, NULL) == -1) {
        
          printf ("Error executing command\n %s\n", command);
          perror ("xfpovray");
          /* sigchld catcher handles the error */
          _exit (255);
       }
            
     }
     else if (pid > 0) {
       /* In parent process */
       
       povray_pid = pid; 
          
     }
 
    /* If verbose output is on, display a Message box, else display 
    ** an Error box if needed.
    */
    if (pov.redirect_on == 1) {

       fl_set_form_position(text_form, 0, 0);
       fl_show_form(text_form, FL_PLACE_POSITION, FL_TRANSIENT, "Messages");
       fl_load_browser(text, STREAM_FILE);

    }
    else {

       if ((fatal_file = fopen(FATAL_FILE, "r")) == NULL) {
          perror("xfpovray: PovCB");
          exit (1);
       }
       else {
          if (fgets(temp, sizeof(temp), fatal_file) != NULL) {
             fl_set_form_position(fatal_form, 0, 0);
             fl_show_form(fatal_form, FL_PLACE_POSITION, FL_TRANSIENT, "Error");
             fl_load_browser(fatal_text, FATAL_FILE);
          }
          fclose(fatal_file);    
       }

    } /* endelse */
    
    /* Display image automatically? */
    if ((auto_view == 1) && (pov.display_during_render != 1))
       ShowImage();

}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* Configuration file options */
void ConfigCB(FL_OBJECT *ob, long data)
{

    int i, submenu;
    const char *save_config_file;
    char temp[256];
    char *t;
    const char *temp2;
         
    submenu = fl_get_menu(ob);

    switch (submenu) {
           case 1: /* Show */
                   /* Config file is shown in the option area, so turn off
                   ** any option radio button that might be selected, then hide
                   ** the current option group
                   */
                   for (i=0; i<4; i++)
                       fl_set_button(opt[i], RELEASED);  
                   fl_hide_object(current_grp);
     
                   current_grp = config_grp;
                   if (loaded_inifile != 1)
                      write_config();  
                   fl_clear_browser(config_browser);
                   fl_load_browser(config_browser, POV_INIFILE);     
    
                   fl_show_object(current_grp);
                   break;
           case 2: /* Save */
                   fl_set_goodies_font(FL_BOLD_STYLE, 12);
                   t = strtok((char *) fl_get_filename(), "."); 
                   if (t == NULL)
                      sprintf(temp, "default.ini");
                   else
                      sprintf(temp, "%s.ini", t);
       
                   save_config_file = fl_show_input(
                       "Enter configuration filename", temp);
    
                   sprintf(temp, "cp %s %s", POV_INIFILE, save_config_file);

                   system(temp);
                   break;
           case 3: /* Load */        
                   temp2 = fl_show_fselector("Enter filename:", 
                       POV_DIR, "*.ini", "");  
                   if (temp2 == NULL) 
                      return;
                   else {
                      loaded_inifile = 1;
                      sprintf(POV_INIFILE, "%s", temp2);         
                   }
                   break;
           default: perror("xfpovray: ConfigCB");
    } /* endswitch */
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* Callback for choice buttons; key off data that is passed in - UUGHH */
void YesNoCB(FL_OBJECT *ob, long data)
{
    switch (data) {
           case 0: pov.display_during_render = fl_get_choice(ob) - 1; 
                   if (pov.display_during_render == 1) {
                      fl_set_object_color(auto_image, 
                          FL_COL1, FL_INACTIVE_COL); 
                      fl_deactivate_object(auto_image);
                   }
                   else {
                      fl_set_object_color(auto_image, FL_COL1, FL_BLACK); 
                      fl_activate_object(auto_image);
                   }
                   break; 
           case 1: switch (fl_get_choice(ob) - 1) {
                          case 0: pov.verbose = 0;
                                  pov.redirect_on = 0;
                                  break;
                          case 1: pov.verbose = 1;
                                  pov.redirect_on = 0;
                                  break;
                          case 2: pov.verbose = 1;
                                  pov.redirect_on = 1;
                                  break;
                          default: perror("xfpovray: YesNoCB");
                   }
                   break;
           case 2: pov.antialias = fl_get_choice(ob) - 1;
                   break;
           case 3: pov.radiosity = fl_get_choice(ob) - 1;
                   break; 
           case 4: auto_view = fl_get_choice(ob) - 1;
                   break;
           case 5: pov.output_to_file = fl_get_choice(ob) - 1;
                   break;
           case 6: pov.version = fl_get_choice(ob);
                   break;
           case 7: pov.remove_bounds = fl_get_choice(ob) - 1;
                   break;
           case 8: pov.split_unions = fl_get_choice(ob) - 1;
                   break;
           case 9: pov.bounding = fl_get_choice(ob) - 1;
                   break;
           case 10: pov.light_buffer = fl_get_choice(ob) - 1;
                   break;
           case 11: pov.vista_buffer = fl_get_choice(ob) - 1;
                   break;
           case 12: pov.aa_sampling_method = fl_get_choice(ob);
                   break;
           case 13: pov.cyclic_animation = fl_get_choice(ob) - 1;
                   break;
           case 14: switch (fl_get_choice(ob) - 1) {
                          case 0: pov.field_rendering = 0;
                                  break;
                          case 1: pov.field_rendering = 1;
                                  pov.odd_field = 0;
                                  break;
                          case 2: pov.field_rendering = 1;
                                  pov.odd_field = 1;
                                  break;
                          default: perror("xfpovray: YesNoCB");
                   } 
                   break;
           case 15: pov.palette = fl_get_choice(ob);
                   break;
           case 16: pov.output_filetype = fl_get_choice(ob);
                   break;
           case 17: pov.jitter = fl_get_choice(ob) - 1;
                   break;
                      
           default: perror("xfpovray: YesNoCB"); /* Error */

    } /* endswitch */
}  


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void QualityCB(FL_OBJECT *ob, long data)
{
    switch (data) {
           case 0: pov.quality = (int) fl_get_slider_value(ob);
                   break;
           case 1: pov.aa_depth = (int) fl_get_slider_value(ob);
                   break;
    }
                   
}
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void QualityResetCB(FL_OBJECT *ob, long data)
{
    switch (data) {
           case 0: fl_set_slider_value(quality_slider, POV_QUALITY);
                   pov.quality = POV_QUALITY;
                   break;
           case 1: fl_set_slider_value(aa_slider, POV_AA_DEPTH);
                   pov.aa_depth = POV_AA_DEPTH;
                   break;
    }

}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void ClearPartialImageCB(FL_OBJECT *ob, long data)
{
    fl_set_input(start_column, "");
    fl_set_input(start_row, "");
    fl_set_input(end_column, "");
    fl_set_input(end_row, "");

}



/* Build the forms for each of the option groups */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void pixmap_group(void)
{
    FL_OBJECT *obj;
    

    pixmap_grp = fl_bgn_group();

       obj = fl_add_box(FL_DOWN_BOX, GRP_X+40-4, GRP_Y+35-4, 320+8, 240+8, "");
           fl_set_object_bw(obj, 4);

       obj = fl_add_pixmap(FL_NORMAL_PIXMAP, GRP_X+40, GRP_Y+35, 320, 240, "");
            fl_set_pixmap_file(obj, STARTUP_PIXMAP);    
 
    fl_end_group(); 
        
}
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void parsing_group(void)
{
    FL_OBJECT *obj;
    int i;    

    parsing_grp = fl_bgn_group();

        obj = fl_add_text(FL_NORMAL_TEXT, 
            GRP_X + 10, GRP_Y + 5, 100, 25, "Library Paths"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
 
        for (i=0; i<5; i++) {
            libpath_input[i] = fl_add_input(FL_NORMAL_INPUT, 
	        GRP_X + 10, GRP_Y + (30*(i+1)), 200, 25, "");                
            if (i == 0)
               fl_set_input(libpath_input[i], POV_LIB_PATH);
            else
               if ((i == 1) && (strcmp(POV_PERSONAL_LIB_PATH, "") != 0)) 
                  fl_set_input(libpath_input[i], POV_PERSONAL_LIB_PATH);
               else
                  fl_set_input(libpath_input[i], "");
        }  

        obj = fl_add_choice(FL_NORMAL_CHOICE, 
            GRP_X + 140, GRP_Y + 180, 70, 25, "Language Version"); 
            fl_addto_choice(obj, "1.0|2.0|3.0");
            fl_set_choice(obj, pov.version);
            fl_set_choice_fontstyle(obj, FL_BOLD_STYLE); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_boxtype(obj, FL_UP_BOX);
            fl_set_object_callback(obj, YesNoCB, 6);

        obj = fl_add_choice(FL_NORMAL_CHOICE, 
            GRP_X + 140, GRP_Y + 210, 70, 25, "Remove Bounds"); 
            fl_addto_choice(obj, "No|Yes");
            fl_set_choice(obj, pov.remove_bounds + 1);
            fl_set_choice_fontstyle(obj, FL_BOLD_STYLE); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_boxtype(obj, FL_UP_BOX);
            fl_set_object_callback(obj, YesNoCB, 7);

        obj = fl_add_choice(FL_NORMAL_CHOICE, 
            GRP_X + 140, GRP_Y + 240, 70, 25, "Split Unions"); 
            fl_addto_choice(obj, "No|Yes");
            fl_set_choice(obj, pov.split_unions + 1);
            fl_set_choice_fontstyle(obj, FL_BOLD_STYLE); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_boxtype(obj, FL_UP_BOX);
            fl_set_object_callback(obj, YesNoCB, 8);
         
    fl_end_group(); 
        
}
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void output_group(void)
{
    FL_OBJECT *obj;
    char temp[10];

    
    output_grp = fl_bgn_group();
    
	obj = fl_add_menu(FL_PULLDOWN_MENU, 
	    GRP_X + 10, GRP_Y + 10, 100, 25, "Image Width");
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_boxtype(obj, FL_UP_BOX);
            fl_set_menu(obj, "50|60|70|80|90|100|150|200|640|800|1024|1280");
            fl_set_object_callback(obj, SetImageSizeCB, 1);

	width_input = fl_add_input(FL_NORMAL_INPUT, 
	    GRP_X + 120, GRP_Y + 10, 50, 25, "");
	    sprintf(temp, "%d", POV_WIDTH);
            fl_set_input(width_input, temp);    

	obj = fl_add_menu(FL_PULLDOWN_MENU, 
	    GRP_X + 10, GRP_Y + 40, 100, 25, "Image Height");
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_boxtype(obj, FL_UP_BOX);
            fl_set_menu(obj, "50|60|70|80|90|100|150|200|480|600|768|1024");
            fl_set_object_callback(obj, SetImageSizeCB, 0);

	height_input = fl_add_input(FL_NORMAL_INPUT, 
	    GRP_X + 120, GRP_Y + 40, 50, 25, "");
	    sprintf(temp, "%d", POV_HEIGHT);
            fl_set_input(height_input, temp);  

	obj = fl_add_menu(FL_PULLDOWN_MENU, 
	    GRP_X + 220, GRP_Y + 10, 100, 25, "Mosaic Start");
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_boxtype(obj, FL_UP_BOX);
            fl_set_menu(obj, "64|32|16|8|4|2| ");
            fl_set_object_callback(obj, SetMosaicCB, 0);
 
        mosaic_start = fl_add_input(FL_NORMAL_INPUT, 
	    GRP_X + 340, GRP_Y + 10, 50, 25, "");
            fl_set_input(mosaic_start, "");  

	obj = fl_add_menu(FL_PULLDOWN_MENU, 
	    GRP_X + 220, GRP_Y + 40, 100, 25, "Mosaic End");
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_boxtype(obj, FL_UP_BOX);
            fl_set_menu(obj, "64|32|16|8|4|2| ");
            fl_set_object_callback(obj, SetMosaicCB, 1);

        mosaic_end = fl_add_input(FL_NORMAL_INPUT, 
	    GRP_X + 340, GRP_Y + 40, 50, 25, "");
            fl_set_input(mosaic_end, "");  
 
        obj = fl_add_text(FL_NORMAL_TEXT,
            GRP_X + 10, GRP_Y + 90, 150, 25, "Crop Image"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);

        obj = fl_add_text(FL_NORMAL_TEXT, 
            GRP_X + 10, GRP_Y + 120, 100, 25, "Start Column"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
 
        start_column = fl_add_input(FL_NORMAL_INPUT, 
	    GRP_X + 100, GRP_Y + 120, 50, 25, "");
            fl_set_input(start_column, "");  

        obj = fl_add_text(FL_NORMAL_TEXT, 
            GRP_X + 170, GRP_Y + 120, 100, 25, "End Column"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
 
        end_column = fl_add_input(FL_NORMAL_INPUT, 
	    GRP_X + 250, GRP_Y + 120, 50, 25, "");
            fl_set_input(end_column, "");  

	obj = fl_add_button(FL_NORMAL_BUTTON, 
	    GRP_X + 320, GRP_Y + 120, 70, 25, "Clear"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_callback(obj, ClearPartialImageCB, 0);


        obj = fl_add_text(FL_NORMAL_TEXT, 
            GRP_X + 10, GRP_Y + 150, 100, 25, "Start Row"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
 
        start_row = fl_add_input(FL_NORMAL_INPUT, 
	    GRP_X + 100, GRP_Y + 150, 50, 25, "");
            fl_set_input(start_row, "");  

        obj = fl_add_text(FL_NORMAL_TEXT, 
            GRP_X + 170, GRP_Y + 150, 100, 25, "End Row"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
 
        end_row = fl_add_input(FL_NORMAL_INPUT, 
	    GRP_X + 250, GRP_Y + 150, 50, 25, "");
            fl_set_input(end_row, "");

        obj = fl_add_choice(FL_NORMAL_CHOICE, 
            GRP_X + 150, GRP_Y + 200, 70, 25, "Display During Render"); 
            fl_addto_choice(obj, "No|Yes");
            fl_set_choice(obj, pov.display_during_render + 1);
            fl_set_choice_fontstyle(obj, FL_BOLD_STYLE); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_boxtype(obj, FL_UP_BOX);
            fl_set_object_callback(obj, YesNoCB, 0);

        obj = fl_add_choice(FL_NORMAL_CHOICE, 
            GRP_X + 320, GRP_Y + 200, 70, 25, "Palette"); 
            fl_addto_choice(obj, "True Color|High Color|Gray Scale");
            fl_set_choice(obj, pov.palette);
            fl_set_choice_fontstyle(obj, FL_BOLD_STYLE); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_boxtype(obj, FL_UP_BOX);
            fl_set_object_callback(obj, YesNoCB, 15);
 
        obj = fl_add_choice(FL_NORMAL_CHOICE, 
            GRP_X + 150, GRP_Y + 230, 70, 25, "Output Image to File"); 
            fl_addto_choice(obj, "No|Yes");
            fl_set_choice(obj, pov.output_to_file + 1);
            fl_set_choice_fontstyle(obj, FL_BOLD_STYLE); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_boxtype(obj, FL_UP_BOX);
            fl_set_object_callback(obj, YesNoCB, 5);

        obj = fl_add_choice(FL_NORMAL_CHOICE, 
            GRP_X + 320, GRP_Y + 230, 70, 25, "File Type"); 
            fl_addto_choice(obj, "TGA-24|RLE|PNG|PPM|System");
            fl_set_choice(obj, pov.output_filetype);
            fl_set_choice_fontstyle(obj, FL_BOLD_STYLE); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_boxtype(obj, FL_UP_BOX);
            fl_set_object_callback(obj, YesNoCB, 16);
 
                 
        obj = fl_add_choice(FL_NORMAL_CHOICE, 
            GRP_X + 150, GRP_Y + 260, 70, 25, "Verbose"); 
            fl_addto_choice(obj, "Off|Low|High");
            fl_set_choice(obj, pov.verbose + 1);
            fl_set_choice_fontstyle(obj, FL_BOLD_STYLE); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_boxtype(obj, FL_UP_BOX);
            fl_set_object_callback(obj, YesNoCB, 1);
 
    fl_end_group(); 
        
}
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void tracing_group(void)
{
    FL_OBJECT *obj;
    char temp[128];
    int bw = 90;
    int bh = 25;

    tracing_grp = fl_bgn_group();

        quality_slider = fl_add_valslider(FL_HOR_SLIDER, 
            GRP_X + 10, GRP_Y + 10, 220, 20, "");
            fl_set_slider_bounds (quality_slider, 1, 11);
            fl_set_slider_value (quality_slider, 9);
            fl_set_slider_step (quality_slider, 1);
            fl_set_slider_precision (quality_slider, 0);
            fl_set_slider_return (quality_slider, 0);
            fl_set_object_lsize (quality_slider, FL_NORMAL_SIZE);
            fl_set_object_callback (quality_slider, QualityCB, 0);
            fl_set_object_lstyle(quality_slider, FL_BOLD_STYLE);

        obj = fl_add_text(FL_NORMAL_TEXT, 
            GRP_X + 270, GRP_Y + 10, 50, 25, "Quality"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);

        obj = fl_add_button (FL_NORMAL_BUTTON, 
            GRP_X + 240, GRP_Y + 10, 20, 20, "@circle");
            fl_set_object_callback (obj, QualityResetCB, 0);
            fl_set_object_lsize(obj, FL_NORMAL_SIZE);
            fl_set_object_bw(obj, +2);

        aa_slider = fl_add_valslider(FL_HOR_SLIDER, 
            GRP_X + 10, GRP_Y + 40, 220, 20, "");
            fl_set_slider_bounds (aa_slider, 1, 9);
            fl_set_slider_value (aa_slider, 3);
            fl_set_slider_step (aa_slider, 1);
            fl_set_slider_precision (aa_slider, 0);
            fl_set_slider_return (aa_slider, 0);
            fl_set_object_lsize (aa_slider, FL_NORMAL_SIZE);
            fl_set_object_callback (aa_slider, QualityCB, 1);
            fl_set_object_lstyle(aa_slider, FL_BOLD_STYLE);

        obj = fl_add_text(FL_NORMAL_TEXT, 
            GRP_X + 270, GRP_Y + 40, 100, 25, "Antialias Depth"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);

        obj = fl_add_button (FL_NORMAL_BUTTON, 
            GRP_X + 240, GRP_Y + 40, 20, 20, "@circle");
            fl_set_object_callback (obj, QualityResetCB, 1);
            fl_set_object_lsize(obj, FL_NORMAL_SIZE);
            fl_set_object_bw(obj, +2);

        obj = fl_add_choice(FL_NORMAL_CHOICE, 
            GRP_X + 80, GRP_Y + 80, bw, bh, "Antialias"); 
            fl_addto_choice(obj, "No|Yes");
            fl_set_choice(obj, pov.antialias + 1);
            fl_set_choice_fontstyle(obj, FL_BOLD_STYLE); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_boxtype(obj, FL_UP_BOX);
            fl_set_object_callback(obj, YesNoCB, 2);

        obj = fl_add_text(FL_NORMAL_TEXT, 
            GRP_X + 170, GRP_Y + 80, 130, 25, "Antialias Threshold"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_lalign(obj, FL_ALIGN_RIGHT);
 
        antialias_threshold = fl_add_input(FL_NORMAL_INPUT, 
	    GRP_X + 305, GRP_Y + 80, 50, 25, "");
	    sprintf(temp, "%f", POV_ANTIALIAS_THRESHOLD);
            fl_set_input(antialias_threshold, temp);
 
        obj = fl_add_choice(FL_NORMAL_CHOICE, 
            GRP_X + 80, GRP_Y + 110, bw, bh, "Method"); 
            fl_addto_choice(obj, "Non-Adaptive|Adaptive");
            fl_set_choice(obj, pov.aa_sampling_method);
            fl_set_choice_fontstyle(obj, FL_BOLD_STYLE); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_boxtype(obj, FL_UP_BOX);
            fl_set_object_callback(obj, YesNoCB, 12);

        obj = fl_add_choice(FL_NORMAL_CHOICE, 
            GRP_X + 80, GRP_Y + 140, bw, bh, "Radiosity"); 
            fl_addto_choice(obj, "No|Yes");
            fl_set_choice(obj, pov.radiosity + 1);
            fl_set_choice_fontstyle(obj, FL_BOLD_STYLE); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_boxtype(obj, FL_UP_BOX);
            fl_set_object_callback(obj, YesNoCB, 3);

        obj = fl_add_choice(FL_NORMAL_CHOICE, 
            GRP_X + 80, GRP_Y + 170, bw, bh, "Bounding"); 
            fl_addto_choice(obj, "No|Yes");
            fl_set_choice(obj, pov.bounding + 1);
            fl_set_choice_fontstyle(obj, FL_BOLD_STYLE); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_boxtype(obj, FL_UP_BOX);
            fl_set_object_callback(obj, YesNoCB, 9);

        obj = fl_add_text(FL_NORMAL_TEXT, 
            GRP_X + 170, GRP_Y + 170, 130, 25, "Bounding Threshold"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_lalign(obj, FL_ALIGN_RIGHT);

        sprintf(temp, "%d", pov.bounding_threshold); 
        bounding_threshold = fl_add_input(FL_NORMAL_INPUT, 
	    GRP_X + 305, GRP_Y + 170, 50, 25, "");
            fl_set_input(bounding_threshold, temp);
 
        obj = fl_add_choice(FL_NORMAL_CHOICE, 
            GRP_X + 80, GRP_Y + 200, bw, bh, "Light Buffer"); 
            fl_addto_choice(obj, "No|Yes");
            fl_set_choice(obj, pov.light_buffer + 1);
            fl_set_choice_fontstyle(obj, FL_BOLD_STYLE); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_boxtype(obj, FL_UP_BOX);
            fl_set_object_callback(obj, YesNoCB, 10);

        obj = fl_add_choice(FL_NORMAL_CHOICE, 
            GRP_X + 80, GRP_Y + 230, bw, bh, "Vista Buffer"); 
            fl_addto_choice(obj, "No|Yes");
            fl_set_choice(obj, pov.vista_buffer + 1);
            fl_set_choice_fontstyle(obj, FL_BOLD_STYLE); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_boxtype(obj, FL_UP_BOX);
            fl_set_object_callback(obj, YesNoCB, 11);

        obj = fl_add_choice(FL_NORMAL_CHOICE, 
            GRP_X + 80, GRP_Y + 260, bw, bh, "Jitter"); 
            fl_addto_choice(obj, "No|Yes");
            fl_set_choice(obj, pov.jitter + 1);
            fl_set_choice_fontstyle(obj, FL_BOLD_STYLE); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_boxtype(obj, FL_UP_BOX);
            fl_set_object_callback(obj, YesNoCB, 17);

        obj = fl_add_text(FL_NORMAL_TEXT, 
            GRP_X + 170, GRP_Y + 260, 130, 25, "Jitter Amount"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_lalign(obj, FL_ALIGN_RIGHT);
 
        jitter_input = fl_add_input(FL_NORMAL_INPUT, 
	    GRP_X + 305, GRP_Y + 260, 50, 25, "");
	    sprintf(temp, "%f", POV_JITTER_AMOUNT);
            fl_set_input(jitter_input, temp);

              
    fl_end_group(); 
        
}
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void animation_group(void)
{
    FL_OBJECT *obj;
    

    animation_grp = fl_bgn_group();

        obj = fl_add_text(FL_NORMAL_TEXT,
            GRP_X + 10, GRP_Y + 10, 200, 25, "Internal Animation Loop"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);

        obj = fl_add_text(FL_NORMAL_TEXT, 
            GRP_X + 20, GRP_Y + 40, 100, 25, "Initial Frame"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
        init_frame_input = fl_add_input(FL_NORMAL_INPUT, 
	    GRP_X + 110, GRP_Y + 40, 50, 25, "");
            fl_set_input(init_frame_input, "1");  

        obj = fl_add_text(FL_NORMAL_TEXT, 
            GRP_X + 180, GRP_Y + 40, 130, 25, "Subset Start Frame"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
        subset_start_frame_input = fl_add_input(FL_NORMAL_INPUT, 
	    GRP_X + 320, GRP_Y + 40, 50, 25, "");
 
        obj = fl_add_text(FL_NORMAL_TEXT, 
            GRP_X + 20, GRP_Y + 70, 100, 25, "Final Frame"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
        final_frame_input = fl_add_input(FL_NORMAL_INPUT, 
	    GRP_X + 110, GRP_Y + 70, 50, 25, "");
            fl_set_input(final_frame_input, "-1");  

        obj = fl_add_text(FL_NORMAL_TEXT, 
            GRP_X + 180, GRP_Y + 70, 130, 25, "Subset End Frame"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
        subset_end_frame_input = fl_add_input(FL_NORMAL_INPUT, 
	    GRP_X + 320, GRP_Y + 70, 50, 25, "");

        obj = fl_add_choice(FL_NORMAL_CHOICE, 
            GRP_X + 300, GRP_Y + 100, 70, 25, "Cyclic Animation"); 
            fl_addto_choice(obj, "Off|On");
            fl_set_choice(obj, pov.cyclic_animation + 1);
            fl_set_choice_fontstyle(obj, FL_BOLD_STYLE); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_boxtype(obj, FL_UP_BOX);
            fl_set_object_callback(obj, YesNoCB, 13);

        obj = fl_add_choice(FL_NORMAL_CHOICE, 
            GRP_X + 300, GRP_Y + 130, 70, 25, "Field Rendering"); 
            fl_addto_choice(obj, "Off|Even|Odd");
            fl_set_choice(obj, pov.field_rendering + 1);
            fl_set_choice_fontstyle(obj, FL_BOLD_STYLE); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
            fl_set_object_boxtype(obj, FL_UP_BOX);
            fl_set_object_callback(obj, YesNoCB, 14);

        obj = fl_add_text(FL_NORMAL_TEXT, 
            GRP_X + 20, GRP_Y + 100, 100, 25, "Initial Clock"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
        initial_clock_input = fl_add_input(FL_NORMAL_INPUT, 
	    GRP_X + 110, GRP_Y + 100, 50, 25, "");
            fl_set_input(initial_clock_input, "0.00");  

        obj = fl_add_text(FL_NORMAL_TEXT, 
            GRP_X + 20, GRP_Y + 130, 100, 25, "Final Clock"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
        final_clock_input = fl_add_input(FL_NORMAL_INPUT, 
	    GRP_X + 110, GRP_Y + 130, 50, 25, "");
            fl_set_input(final_clock_input, "1.00");  



        obj = fl_add_text(FL_NORMAL_TEXT, 
            GRP_X + 10, GRP_Y + 170, 200, 25, "External Animation Loop"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);

        obj = fl_add_text(FL_NORMAL_TEXT, 
            GRP_X + 20, GRP_Y + 200, 100, 25, "Clock"); 
            fl_set_object_lstyle(obj, FL_BOLD_STYLE);
        clock_input = fl_add_input(FL_NORMAL_INPUT, 
	    GRP_X + 110, GRP_Y + 200, 50, 25, "");
            fl_set_input(clock_input, "0.00");

    fl_end_group(); 
        
}
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void config_group(void)
{
    FL_OBJECT *obj;

    config_grp = fl_bgn_group();
   
        config_browser = fl_add_browser(FL_NORMAL_BROWSER, 
            GRP_X, GRP_Y, GRP_W, GRP_H, POV_INIFILE); 

        fl_load_browser(config_browser, POV_INIFILE);
         
    fl_end_group(); 

}
 


/* Miscellaneous functions */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* Function to write the povray configuration file */
void write_config(void)
{
    char temp[256];    
    FILE *options_file;
    int i;
   
   
           
    if ((options_file = fopen(POV_INIFILE, "w")) == NULL) {
       perror("xfpovray: write_config");
       exit (1);
    }



    fprintf(options_file, "version=%3.1f\n", (float) pov.version);    

    fprintf(options_file, "input_file_name=%s\n", pov.input_file_name);
    
    /* Library path */
    for (i=0; i<5; i++) {
        if (strcmp(fl_get_input(libpath_input[i]), "") != 0)
           fprintf(options_file, "library_path=%s\n", 
               fl_get_input(libpath_input[i]));
    }
               
    /* Image height and width. 
    ** Must parse the input boxes in case the user 
    ** made an entry by clicking in the input box instead
    ** of using the pulldown menu.
    */
    if (strcmp(fl_get_input(height_input), "") == 0) {
       sprintf(temp, "%d", POV_HEIGHT);
       fl_set_input(height_input, temp);
       pov.height = POV_HEIGHT;
    }
    else
       pov.height = atoi(fl_get_input(height_input));

    if (strcmp(fl_get_input(width_input), "") == 0) {
       sprintf(temp, "%d", POV_WIDTH);
       fl_set_input(width_input, temp);
       pov.width = POV_WIDTH;
    }
    else
       pov.width = atoi(fl_get_input(width_input)); 
       
    fprintf(options_file, "height=%d\n", pov.height); 
    fprintf(options_file, "width=%d\n", pov.width); 

    fprintf(options_file, "verbose=%d\n", pov.verbose);

    /* Redirection options.                   
    ** Turn off redirect to console; redirect to file if selected,
    ** then display file in message browser.
    */
    fprintf(options_file, "all_console=%d\n", 0);
    if (pov.redirect_on == 1)
       fprintf(options_file, "all_file=%s\n", STREAM_FILE); 
    else
       fprintf(options_file, "fatal_file=%s\n", FATAL_FILE);
 
    /* Partial image height and width; parse the input boxes */ 
    if (strcmp(fl_get_input(start_column), "") != 0)
       fprintf(options_file, "start_column=%s\n", fl_get_input(start_column)); 
    if (strcmp(fl_get_input(end_column), "") != 0)
       fprintf(options_file, "end_column=%s\n", fl_get_input(end_column)); 
    if (strcmp(fl_get_input(start_row), "") != 0)
       fprintf(options_file, "start_row=%s\n", fl_get_input(start_row)); 
    if (strcmp(fl_get_input(end_row), "") != 0)
       fprintf(options_file, "end_row=%s\n", fl_get_input(end_row)); 

    /* Mosaic preview; parse the input boxes */
    if (strcmp(fl_get_input(mosaic_start), "") != 0)
       fprintf(options_file, "preview_start_size=%s\n", 
               fl_get_input(mosaic_start)); 
    if (strcmp(fl_get_input(mosaic_end), "") != 0)
       fprintf(options_file, "preview_end_size=%s\n", 
               fl_get_input(mosaic_end)); 
 
  
    fprintf(options_file, "quality=%d\n", pov.quality); 

    fprintf(options_file, "antialias=%d\n", pov.antialias); 
    fprintf(options_file, "antialias_depth=%d\n", pov.aa_depth); 
    fprintf(options_file, "sampling_method=%d\n", pov.aa_sampling_method); 
    fprintf(options_file, "antialias_threshold=%s\n", 
        fl_get_input(antialias_threshold)); 

    fprintf(options_file, "radiosity=%d\n", pov.radiosity); 

    fprintf(options_file, "display=%d\n", pov.display_during_render);
        
    /* If displaying during a render, pause to view image */    
    if (pov.display_during_render == 1)
       fprintf(options_file, "pause_when_done=%d\n", 1); 

    fprintf(options_file, "output_to_file=%d\n", pov.output_to_file);

    fprintf(options_file, "remove_bounds=%d\n", pov.remove_bounds);

    fprintf(options_file, "split_unions=%d\n", pov.split_unions); 

    fprintf(options_file, "bounding=%d\n", pov.bounding);
    
    if (strcmp(fl_get_input(bounding_threshold), "") != 0)
       pov.bounding_threshold = atoi(fl_get_input(bounding_threshold));
    else 
       pov.bounding_threshold = POV_BOUNDING_THRESHOLD; 
    fprintf(options_file, "bounding_threshold=%d\n", pov.bounding_threshold); 

    fprintf(options_file, "light_buffer=%d\n", pov.light_buffer);

    fprintf(options_file, "vista_buffer=%d\n", pov.vista_buffer);

    fprintf(options_file, "clock=%s\n", fl_get_input(clock_input)); 

    fprintf(options_file, "initial_clock=%s\n", 
        fl_get_input(initial_clock_input)); 

    fprintf(options_file, "final_clock=%s\n", 
        fl_get_input(final_clock_input)); 
 
    pov.initial_frame = atoi(fl_get_input(init_frame_input));
    fprintf(options_file, "initial_frame=%d\n", pov.initial_frame);
      
    pov.final_frame = atoi(fl_get_input(final_frame_input));
    fprintf(options_file, "final_frame=%d\n", pov.final_frame);

    if (strcmp(fl_get_input(subset_start_frame_input), "") != 0) {
       fprintf(options_file, "subset_start_frame=%s\n", 
           fl_get_input(subset_start_frame_input));
    }
    if (strcmp(fl_get_input(subset_end_frame_input), "") != 0) {
       fprintf(options_file, "subset_end_frame=%s\n", 
           fl_get_input(subset_end_frame_input));
    }
       
    fprintf(options_file, "cyclic_animation=%d\n", pov.cyclic_animation);

    fprintf(options_file, "field_render=%d\n", pov.field_rendering);
    if (pov.field_rendering == 1)
       fprintf(options_file, "odd_field=%d\n", pov.odd_field);

    switch (pov.palette) {
           case 1: fprintf(options_file, "palette=T\n");
                   break;
           case 2: fprintf(options_file, "palette=H\n");
                   break;
           case 3: fprintf(options_file, "palette=G\n");
                   break;
    }           

    switch (pov.output_filetype) {
           case 1: fprintf(options_file, "output_file_type=T\n");
                   break;
           case 2: fprintf(options_file, "output_file_type=C\n");
                   break;
           case 3: fprintf(options_file, "output_file_type=N\n");
                   break;
           case 4: fprintf(options_file, "output_file_type=P\n");
                   break;
           case 5: fprintf(options_file, "output_file_type=S\n");
                   break;
    }           

    fprintf(options_file, "jitter=%d\n", pov.jitter); 

    fprintf(options_file, "jitter_amount=%s\n", fl_get_input(jitter_input)); 

    
    fclose(options_file);

 
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* Display the image after rendering.  Some OS do not allow
** display during rendering :-(
*/
void ShowImage(void)
{
    char viewit[256];
    char imgfile[256];
    char temp[256];
    

    strcpy(temp, pov.input_file_name);
    sprintf(imgfile, "%s", strtok(temp, "."));
    strcat(imgfile, ".tga");
    
    sprintf(viewit, "%s %s &", DISPLAY_COMMAND, imgfile);

    system(viewit);

}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
int MoveCB(XEvent *ev, void *data)
{
    static FL_Coord dx = 1, dy = 1;
    FL_Coord x, y, w, h;  


    fl_get_object_geometry(tc, &x, &y, &w, &h);

    if ( (x + dx) < 0 || (x + w + dx) >= tc->form->w)
       dx = -dx;
    
    if ( (y + dy) < 0 || (y + (h+140) + dy) >= tc->form->h)
       dy = -dy;
    
    x += dx;
    y += dy;

    fl_set_object_position(tc, x ,y);
       
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void ContinueCB(FL_OBJECT *ob, long data)
{
    int r, val;
    int num_colors = 10;
    int colors[10] = {FL_RED,
                      FL_GREEN,
                      FL_YELLOW,
                      FL_BLUE,
                      FL_MAGENTA,
                      FL_SLATEBLUE,
                      FL_ORCHID,
                      FL_DARKGOLD,
                      FL_TOMATO,
                      FL_WHITE};
    
    fl_set_timer(tim, TIME);
    fl_set_object_callback(tim, ContinueCB, 0);

    /* Set a random color for tc */
    srand(time(0));
    r = rand();
    val = (int) ( (r / (float) RAND_MAX) * (num_colors) ); 
       
    fl_set_object_lcol(tc, colors[val]);

}
